'use strict';

exports.__esModule = true;

var _utils = require('./utils');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var ASTObject = function () {
  function ASTObject(obj, expressions, composesCount, t) {
    _classCallCheck(this, ASTObject);

    this.obj = obj;
    this.expressions = expressions;
    this.composesCount = composesCount;
    this.t = t;
  }

  ASTObject.prototype.toAST = function toAST() {
    var obj = this.obj,
        t = this.t;


    var props = [];
    for (var key in obj) {
      var rawValue = obj[key];

      var _objKeyToAst = this.objKeyToAst(key),
          _computed = _objKeyToAst.computed,
          _composes = _objKeyToAst.composes,
          keyAST = _objKeyToAst.ast;

      var valueAST = void 0;
      if (_composes) {
        // valueAST = t.arrayExpression(expressions.slice(0, composesCount))
        continue;
      } else {
        valueAST = this.objValueToAst(rawValue);
      }

      props.push(t.objectProperty(keyAST, valueAST, _computed));
    }
    return t.objectExpression(props);
  };

  ASTObject.prototype.objKeyToAst = function objKeyToAst(key) {
    var t = this.t;

    var matches = this.getDynamicMatches(key);

    if (matches.length) {
      return {
        computed: true,
        composes: key === 'composes',
        ast: this.replacePlaceholdersWithExpressions(matches, key)
      };
    }

    return {
      computed: false,
      composes: key === 'composes',
      ast: t.stringLiteral(key)
    };
  };

  ASTObject.prototype.objValueToAst = function objValueToAst(value) {
    var _this = this;

    var expressions = this.expressions,
        composesCount = this.composesCount,
        t = this.t;


    if (typeof value === 'string') {
      var matches = this.getDynamicMatches(value);
      if (matches.length) {
        return this.replacePlaceholdersWithExpressions(matches, value);
      }
      return t.stringLiteral(value);
    } else if (Array.isArray(value)) {
      // should never hit here
      return t.arrayExpression(value.map(function (v) {
        return _this.objValueToAst(v);
      }));
    }

    var obj = new this.constructor(value, expressions, composesCount, t);
    return obj.toAST();
  };

  ASTObject.prototype.getDynamicMatches = function getDynamicMatches(str) {
    var re = /xxx(\d+)xxx/gm;
    var match = void 0;
    var matches = [];
    while ((match = re.exec(str)) !== null) {
      matches.push({
        value: match[0],
        p1: parseInt(match[1], 10),
        index: match.index
      });
    }

    return matches;
  };

  ASTObject.prototype.replacePlaceholdersWithExpressions = function replacePlaceholdersWithExpressions(matches, str) {
    var expressions = this.expressions,
        composesCount = this.composesCount,
        t = this.t;


    var templateElements = [];
    var templateExpressions = [];
    var cursor = 0;
    // not sure how to detect when to add 'px'
    // let hasSingleInterpolation = false
    (0, _utils.forEach)(matches, function (_ref, i) {
      var value = _ref.value,
          p1 = _ref.p1,
          index = _ref.index;

      var preMatch = str.substring(cursor, index);
      cursor = cursor + preMatch.length + value.length;
      if (preMatch) {
        templateElements.push(t.templateElement({ raw: preMatch, cooked: preMatch }));
      } else if (i === 0) {
        templateElements.push(t.templateElement({ raw: '', cooked: '' }));
      }
      // if (value === str) {
      // hasSingleInterpolation = true
      // }

      templateExpressions.push(expressions ? expressions[p1 - composesCount] : t.identifier('x' + (p1 - composesCount)));
      if (i === matches.length - 1) {
        templateElements.push(t.templateElement({
          raw: str.substring(index + value.length),
          cooked: str.substring(index + value.length)
        }, true));
      }
    });
    // if (hasSingleInterpolation) {
    //   return templateExpressions[0]
    // }
    return t.templateLiteral(templateElements, templateExpressions);
  };

  ASTObject.fromAST = function fromAST(astObj, t) {
    function isLiteral(value) {
      return t.isStringLiteral(value) || t.isNumericLiteral(value) || t.isBooleanLiteral(value);
    }

    var expressions = [];

    function replaceExpressionsWithPlaceholders(node) {
      if (t.isArrayExpression(node)) {
        return node.elements.map(replaceExpressionsWithPlaceholders);
      } else if (isLiteral(node)) {
        return node.value;
      } else if (t.isTemplateLiteral(node)) {
        var strs = node.quasis.map(function (x) {
          return x.value.cooked;
        });
        var exprs = node.expressions;
        return (0, _utils.reduce)(strs, function (arr, str, i) {
          arr.push(str);
          if (i !== strs.length - 1) {
            expressions.push(exprs[i]);
            arr.push('xxx' + (expressions.length - 1) + 'xxx');
          }
          return arr;
        }, []).join('').trim();
      }

      expressions.push(node);
      return 'xxx' + (expressions.length - 1) + 'xxx';
    }

    function toObj(astObj) {
      var obj = {};

      (0, _utils.forEach)(astObj.properties, function (property) {
        // nested objects
        var key = void 0;
        if (property.computed) {
          key = replaceExpressionsWithPlaceholders(property.key);
        } else {
          key = t.isIdentifier(property.key) ? property.key.name : property.key.value;
        }
        if (t.isObjectExpression(property.value)) {
          obj[key] = toObj(property.value);
        } else {
          obj[key] = replaceExpressionsWithPlaceholders(property.value);
        }
      });

      return obj;
    }

    var obj = toObj(astObj);
    return new ASTObject(obj, expressions, 0, // composesCount: we should support this,
    t);
  };

  return ASTObject;
}();

exports.default = ASTObject;